;;;; srfi-76.scm
;
; Copyright (C) Michael Sperber (2005). All Rights Reserved. 
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation files
; (the "Software"), to deal in the Software without restriction,
; including without limitation the rights to use, copy, modify, merge,
; publish, distribute, sublicense, and/or sell copies of the Software,
; and to permit persons to whom the Software is furnished to do so,
; subject to the following conditions:
; 
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.


(declare
  (fixnum) )


(module srfi-76-base
    (make-record-type-descriptor
     actual-record-type-descriptor
     record-type-descriptor?
     record-constructor record-predicate
     record-accessor record-mutator
     record-type-name record-type-parent record-type-sealed? record-type-uid
     record-type-field-names record-type-opaque? record-field-mutable?
     record? record-type-descriptor
     (define-type/explicit define-type-1 do-append make-record-type 
       define-record-field real-record-type-rtd record-type-args-proc)
     syntax->list
     (type-descriptor real-record-type-rtd) )

(define (syntax->list x)
  (syntax-case x ()
    (() '())
    ((h . t) (cons #'h
		   (syntax->list #'t)))))

(define-record-type :field-spec
  (make-field-spec mutable? name)
  field-spec?
  (mutable? field-spec-mutable?)
  (name field-spec-name))

(define (field-spec=? spec-1 spec-2)
  (and (eq? (field-spec-mutable? spec-1)
	    (field-spec-mutable? spec-2))
       (eq? (field-spec-name spec-1)
	    (field-spec-name spec-2))))

(define-record-type :record-type-descriptor
  (really-make-record-type-descriptor name parent uid sealed? opaque? field-specs)
  record-type-descriptor?
  (name record-type-name)
  (parent record-type-parent)
  ;; this is #f in the generative case
  (uid record-type-uid)
  (sealed? record-type-sealed?)
  (opaque? record-type-opaque?)
  (field-specs record-type-field-specs))

(define (record-type-descriptor=? rtd-1 rtd-2)
  (and (eq? (record-type-parent rtd-1) (record-type-parent rtd-2))
       (eq? (record-type-uid rtd-1) (record-type-uid rtd-2))
       (every field-spec=?
	      (record-type-field-specs rtd-1)
	      (record-type-field-specs rtd-2))))

(define (uid->record-type-descriptor uid)
  (find (lambda (rtd)
	  (eq? (record-type-uid rtd) uid))
	*nongenerative-record-types*))

(define *nongenerative-record-types* '())

(define (make-record-type-descriptor name parent uid sealed? opaque? field-specs)
  (if (and parent
	   (record-type-sealed? parent))
      (error "can't extend a sealed parent class" parent))
  (if (and parent
	   (not (record-type-uid parent)) ; parent generative
	   uid)			  ; ... but this one is non-generative
      (error "a generative type can only be extended to give a generative type"))
  (let ((opaque? (if parent
		     (or (record-type-opaque? parent)
			 opaque?)
		     opaque?))
	(field-specs (map parse-field-spec field-specs)))
    (let ((rtd (really-make-record-type-descriptor name parent uid sealed? opaque? field-specs)))
      (if uid
	  (cond
	   ((uid->record-type-descriptor uid)
	    => (lambda (old-rtd)
		 (if (record-type-descriptor=? rtd old-rtd)
		     old-rtd
		     (error "mismatched nongenerative record types with identical uids"
			    old-rtd rtd))))
	   (else
	    (set! *nongenerative-record-types* 
		  (cons rtd *nongenerative-record-types*))
	    rtd))
	  rtd))))

(define (parse-field-spec spec)
  (apply (lambda (mutability name)
	   (make-field-spec
	    (case mutability
	      ((mutable) #t)
	      ((immutable) #f)
	      (else (error "field spec with invalid mutability specification" spec)))
	    name))
	 spec))

(define (record-type-field-names rtd)
  (map field-spec-name (record-type-field-specs rtd)))

(define (field-count rtd)
  (let loop ((rtd rtd)
	     (count 0))
    (if (not rtd)
	count
	(loop (record-type-parent rtd)
	      (+ count (length (record-type-field-specs rtd)))))))
	 

(define (field-index rtd field)
  (+ (field-count (record-type-parent rtd))
     (cond
      ((list-index (cut eq? field <>) (record-type-field-names rtd)))
      (else
       (error "record type has no such field" rtd field)))))

(define-record-type :record
  (really-make-record rtd components)
  any-record?
  (rtd actual-record-type-descriptor)
  (components record-components))

(define (record? r)
  (and (any-record? r)
       (not (record-type-opaque? (actual-record-type-descriptor r)))))

(define (make-record rtd component-count)
  (really-make-record rtd
		      (make-vector component-count)))

(define (record-type-descriptor r)
  (let ((rtd (actual-record-type-descriptor r)))
    (if (record-type-opaque? rtd)
	(error "tried to obtain record-type descriptor from object of unknown type" r)
	rtd)))

(define (record-ref record index)
  (vector-ref (record-components record) index))

(define (record-set! record index val)
  (vector-set! (record-components record) index val))

(define (record-constructor rtd)
    (let ((component-count (field-count rtd)))
      (lambda components
	(if (not (= (length components) component-count))
	    (error "wrong number of arguments to record constructor" rtd components))
	(let ((r (make-record rtd component-count)))
	  (let loop ((i 0)
		     (components components))
	    (if (not (null? components))
		(begin
		  (record-set! r i (car components))
		  (loop (+ 1 i) (cdr components)))))
	  r))))

; does RTD-1 represent an ancestor of RTD-2?
(define (rtd-ancestor? rtd-1 rtd-2)
  (let loop ((rtd-2 rtd-2))
    (or (eq? rtd-1 rtd-2)
	(and rtd-2
	     (loop (record-type-parent rtd-2))))))

(define (record-with-rtd? obj rtd)
  (and (any-record? obj)
       (rtd-ancestor? rtd (actual-record-type-descriptor obj))))

(define (record-predicate rtd)
  (cut record-with-rtd? <> rtd))

(define (record-accessor rtd field-id)
  (let ((index (field-id-index rtd field-id)))
    (lambda (thing)
      (if (record-with-rtd? thing rtd)
	  (record-ref thing index)
	  (error "accessor applied to bad value" rtd field-id thing)))))

(define (record-mutator rtd field-id)
  (if (not (field-spec-mutable? (field-spec-ref rtd field-id)))
      (error "record-mutator called on immutable field" rtd field-id))
  (let ((index (field-id-index rtd field-id)))
    (lambda (thing val)
      (if (record-with-rtd? thing rtd)
	  (record-set! thing index val)
	  (error "mutator applied to bad value" rtd field-id thing val)))))

; A FIELD-ID may be either an index or a symbol, which needs to refer
; to a field in RTD itself.
(define (field-id-index rtd field-id)
  (if (integer? field-id)
      field-id
      (field-index rtd field-id)))

(define (record-field-mutable? rtd field-id)
  (field-spec-mutable? (field-spec-ref rtd field-id)))

(define (field-spec-ref rtd field-id)
  
  (define (nth rtd index)
    (list-ref (record-type-field-specs rtd) index))
  
  (cond
   ((number? field-id)
    (nth rtd (field-index rtd field-id)))
   ((symbol? field-id)
    (let loop ((rtd rtd))
      (cond
       ((not rtd)
	(error "invalid field spec" rtd field-id))
       ((list-index (cut eq? field-id <>) (record-type-field-names rtd))
	=> (cut nth rtd <>))
       (else
	(loop (record-type-parent rtd))))))
   (else
    (error "invalid field spec" rtd field-id))))

(define-record-type :record-type
  (make-record-type rtd args-proc)
  record-type?
  (rtd real-record-type-rtd)
  ;; Accepts arguments to this record type's constructor.  Returns
  ;; list of field values for this record type.
  (args-proc record-type-args-proc))

(define-syntax type-descriptor
  (syntax-rules ()
    ((type-descriptor ?record-name)
     (real-record-type-rtd ?record-name))))

(define-syntax define-type/explicit
  (syntax-rules ()
    ((define-type (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       #f #f ()		       ; parent, parent rtd, parent init exprs
       #f				; sealed?
       #f				; opaque?
       ()				; fields
       #f				; nongenerative uid
       values				; INIT! proc
       ?clause ...))))

(define-syntax define-type-1
  (syntax-rules (parent sealed nongenerative init! opaque fields)
    ;; find PARENT clause
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?fields-clause ?nongenerative-uid ?init-proc
       (parent ?parent-name ?expr ...)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent-name (type-descriptor ?parent-name)
       (?expr ...) ?sealed? ?opaque? ?fields-clause ?nongenerative-uid ?init-proc
       ?clause ...))

    ;; find SEALED clause
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?nongenerative-uid ?init-proc
       (sealed)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       #t ?opaque? ?field-specs ?nongenerative-uid ?init-proc
       ?clause ...))

    ;; find OPAQUE clause
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?nongenerative-uid ?init-proc
       (opaque)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs 
       ?sealed? #t ?field-specs ?nongenerative-uid ?init-proc
       ?clause ...))

    ;; parse FIELDS clause

    ;; base case
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       (fields)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       ?clause ...))

    ;; missing init expression
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       (fields (?field-name ?procs) ?rest ...)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       (fields (?field-name ?procs ?field-name) ?rest ...)
       ?clause ...))
     
     ;; complete spec
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       (fields (?field-name (?accessor) ?init) ?rest ...)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque?
       (?field-spec ... (immutable ?field-name (?accessor) ?init)) 
       ?nongenerative-uid  ?init-proc
       (fields ?rest ...)
       ?clause ...))

    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? (?field-spec ...) ?nongenerative-uid  ?init-proc
       (fields (?field-name (?accessor ?mutator) ?init) ?rest ...)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque?
       (?field-spec ... (mutable ?field-name (?accessor ?mutator) ?init))
       ?nongenerative-uid  ?init-proc
       (fields ?rest ...)
       ?clause ...))

    ;; find NONGENERATIVE clause
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?nongenerative-uid  ?init-proc
       (nongenerative ?uid)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?uid
       ?init-proc
       ?clause ...))

    ;; find INIT! clause
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?nongenerative-uid  ?init-proc
       (init! (?r) ?body)
       ?clause ...)
     (define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd ?parent-init-exprs
       ?sealed? ?opaque? ?field-specs ?nongenerative-uid
       (lambda (?r) ?body)
       ?clause ...))

    ;; generate code
    ((define-type-1 (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?parent ?parent-rtd (?parent-init-expr ...)
       ?sealed? ?opaque?
       ((?mutability ?field-name ?procs ?init-expr) ...)
       ?nongenerative-uid
       ?init-proc)

     (begin
       ;; where we need LETREC* semantics if this is to work internally
       (define $rtd
	 (make-record-type-descriptor '?record-name
				      ?parent-rtd
				      '?nongenerative-uid
				      ?sealed?
				      ?opaque?
				      '((?mutability ?field-name) ...)))

       (define $args-proc
	 (if ?parent-rtd
	     (let ((parent-args-proc (record-type-args-proc ?parent)))
	       (lambda (child-field-values . ?formals)
		 (parent-args-proc
		  (do-append (?init-expr ...) child-field-values)
		  ?parent-init-expr ...)))
	     (lambda (child-field-values . ?formals)
	       (do-append (?init-expr ...) child-field-values))))

       (define ?record-name (make-record-type $rtd $args-proc))

       (define ?constructor-name
	 (let ((make (record-constructor $rtd)))
	   (if ?parent-rtd
	       (let ((parent-args-proc (record-type-args-proc ?parent)))
		 (lambda ?formals
		   (let ((r (apply make (parent-args-proc (list ?init-expr ...)
							  ?parent-init-expr ...))))
		     (?init-proc r)
		     r)))
	       (lambda ?formals
		 (let ((r (make ?parent-init-expr ... ?init-expr ... )))
		   (?init-proc r)
		   r)))))
       
       (define ?predicate-name
	 (record-predicate $rtd))

       (define-record-field $rtd ?field-name ?procs)
       ...))))

(define-syntax define-record-field
  (syntax-rules ()
    ((define-record-field ?rtd
       ?field-name (?accessor-name))
     (define ?accessor-name (record-accessor ?rtd '?field-name)))
    ((define-record-field ?rtd
       ?field-name (?accessor-name ?mutator-name))
     (begin
       (define ?accessor-name (record-accessor ?rtd '?field-name))
       (define ?mutator-name (record-mutator ?rtd '?field-name))))))

(define-syntax do-append
  (syntax-rules ()
    ((do-append (?elem1 ?elems ...) ?tail)
     (cons ?elem1 (do-append (?elems ...) ?tail)))
    ((do-append () ?tail) ?tail)))
	     
)

(module srfi-76
    (make-record-type-descriptor
     record-type-descriptor?
     record-constructor record-predicate
     record-accessor record-mutator
     record-type-name record-type-parent record-type-sealed? record-type-uid
     record-type-field-names record-type-opaque? record-field-mutable?
     record? record-type-descriptor
     (define-type define-type-1 define-type-2 process-fields-clause syntax->list)
     type-descriptor)

  (import srfi-76-base)

(define-syntax define-type
  (syntax-rules ()
    ((define-type (?record-name ?constructor-name ?predicate-name)
       ?formals
       ?clause ...)
     (define-type-1 ?record-name (?record-name ?constructor-name ?predicate-name)
       ?formals
       ()
       ?clause ...))
    ((define-type ?record-name
       ?formals
       ?clause ...)
     (define-type-1 ?record-name ?record-name
       ?formals
       ()
       ?clause ...))))

(define-syntax define-type-1
  (syntax-rules (fields)

    ;; find FIELDS clause
    ((define-type-1 ?record-name ?record-name-spec
       ?formals
       (?simple-clause ...)
       (fields ?field-spec ...)
       ?clause ...)
     (process-fields-clause (fields ?field-spec ...)
			    ?record-name ?record-name-spec
			    ?formals
			    (?simple-clause ...)
			    ?clause ...))

    ;; collect all other clauses
    ((define-type-1 ?record-name ?record-name-spec
       ?formals
       (?simple-clause ...)
       ?clause0
       ?clause ...)
     (define-type-1 ?record-name ?record-name-spec
       ?formals
       (?simple-clause ... ?clause0)
       ?clause ...))

    ;; pass it on
    ((define-type-1 ?record-name ?record-name-spec
       ?formals
       (?simple-clause ...))

     (define-type-2 ?record-name ?record-name-spec
       ?formals
       (?simple-clause ...)))))

(define-syntax define-type-2
  (lambda (form)
    (syntax-case form ()
      ((_ ?record-name (?record-name-2 ?constructor-name ?predicate-name)
       ?formals
       (?simple-clause ...))

       (syntax
	(begin
	  (define-type/explicit (?record-name ?constructor-name ?predicate-name)
	    ?formals
	    ?simple-clause ...))))

      ((_ ?record-name ?record-name-2
       ?formals
       (?simple-clause ...))

       (with-syntax ((?constructor-name
		      (datum->syntax-object (syntax ?record-name)
					    (string->symbol
					     (string-append "make-"
							    (symbol->string
							     (syntax-object->datum
							      (syntax ?record-name)))))))
		     (?predicate-name
		      (datum->syntax-object (syntax ?record-name)
					    (string->symbol
					     (string-append (symbol->string
							     (syntax-object->datum
							      (syntax ?record-name)))
							    "?")))))
		     
					    
       (syntax
	(define-type-2 ?record-name (?record-name ?constructor-name ?predicate-name)
	  ?formals
	  (?simple-clause ...))))))))

(define-syntax process-fields-clause
  (lambda (form)
    (syntax-case form (fields mutable immutable)
      ((_ (fields ?field-clause ...)
	  ?record-name ?record-name-spec
	  ?formals
	  (?simple-clause ...)
	  ?clause ...)

       (let ((record-name (symbol->string (syntax-object->datum (syntax ?record-name)))))
	 (with-syntax
	     (((?simple-field ...)
	       (map (lambda (clause)
		      (syntax-case clause (mutable immutable)
			((?field-name immutable ?init ...)
			 (with-syntax ((?accessor-name
					(datum->syntax-object
					 (syntax ?field-name)
					 (string->symbol
					  (string-append record-name "-"
							 (symbol->string
							  (syntax-object->datum
							   (syntax ?field-name))))))))
			   (syntax
			    (?field-name (?accessor-name) ?init ...))))
			((?field-name mutable ?init ...)
			 (with-syntax ((?accessor-name
					(datum->syntax-object
					 (syntax ?field-name)
					 (string->symbol
					  (string-append record-name "-"
							 (symbol->string
							  (syntax-object->datum
							   (syntax ?field-name)))))))
				       (?mutator-name
					(datum->syntax-object
					 (syntax ?field-name)
					 (string->symbol
					  (string-append record-name "-"
							 (symbol->string
							  (syntax-object->datum
							   (syntax ?field-name)))
							 "-set!")))))
			   (syntax
			    ( ?field-name (?accessor-name ?mutator-name) ?init ...))))
			(?clause
			 clause)))
		    (syntax->list (syntax (?field-clause ...))))))
	   (syntax
	    (define-type-1 
	      ?record-name ?record-name-spec
	      ?formals
	      (?simple-clause ... (fields ?simple-field ...))
	      ?clause ...))))))))

(define-record-printer (:record record port)
  (fprintf port "#<~a>" (record-type-name (actual-record-type-descriptor record))) )

)
